import argparse
import fnmatch
import shutil
import threading
import traceback
from copy import deepcopy
from http.server import HTTPServer
from pathlib import Path

from kart.utils import KartMap, KartObserver, KartRequestHandler, merge_dicts


class Kart:
    """Main Kart class"""

    def __init__(
        self,
        miners: list = [],
        content_modifiers: list = [],
        mappers: list = [],
        map_modifiers: list = [],
        renderers: list = [],
        config: dict = {},
        build_location: str = "_site",
    ):
        self.miners = miners
        self.content_modifiers = content_modifiers
        self.mappers = mappers
        self.map_modifiers = map_modifiers
        self.renderers = renderers
        self.config = config
        self.build_location = Path(build_location)
        self.lock = threading.Lock()

    def check_config(self):
        """Checks if the config has all the necessary fields and sets them to default values if not"""
        default = {
            "name": "Example",
            "site_url": "https://example.org",
            "pagination": {"per_page": 5, "skip": 0},
            "code_highlighting": {"style": "default", "noclasses": True},
            "timezone": "UTC",
            "serving": False,
        }
        merge_dicts(self.config, default)

    def mine_data(self, start: bool = True):
        """Calls miners and content modifiers"""
        self.site = {}
        for miner in self.miners:
            if start:
                miner.read_data(self.config)
            self.site.update(miner.collect(self.config))
        for modifier in self.content_modifiers:
            modifier.modify(self.config, self.site)

    def create_map(self):
        """Calls mappers and map modifiers"""
        self.map = KartMap(site_url=self.config["site_url"])
        for mapper in self.mappers:
            self.map.update(mapper.map(self.config, self.site))
        for modifier in self.map_modifiers:
            modifier.modify(self.config, self.site, self.map)

    def write(self):
        """Calls renderers"""
        for renderer in self.renderers:
            renderer.render(self.config, self.site, self.map, self.build_location)

    def build(self):
        """Build the entire site"""
        self.check_config()
        self.mine_data()
        self.create_map()
        shutil.rmtree(self.build_location, ignore_errors=True)
        self.build_location.mkdir(parents=True, exist_ok=True)
        self.write()

    # _site and _map are set and retrieved with a threading lock to prevent data races
    # _site and _map are set only when the creation of the map has finished
    # therefore it is not possible to access only partial data,
    # preventing errors when serving the site during development

    def update_data(self):
        """Update the site data after a file has been changed"""
        self.mine_data(False)
        self.create_map()
        _site = deepcopy(self.site)
        _map = deepcopy(self.map)
        _urls = {}
        _regexes = {}
        for slug, page in self.map.items():
            _urls[page["url"]] = slug
            if "*" in page["url"] or "?" in page["url"]:
                _regexes[page["url"]] = slug
        with self.lock:
            self._site = _site
            self._map = _map
            self._urls = _urls
            self._regexes = _regexes

    def serve_page(self, handler, url: str):
        """Serve a single page"""
        with self.lock:
            site_map = self._map
            urls = self._urls
            site = self._site
            regexes = self._regexes
        if url in urls:
            page = site_map[urls[url]]
        else:
            try:
                pattern = next((x for x in regexes if fnmatch.fnmatch(url, x)))
                page = site_map[regexes[pattern]]
            except StopIteration:
                page = None
        if page:
            renderer = self.renderer_dict[page["renderer"]]
            renderer.serve(handler, page, self.config, site, site_map)

    def serve(self, port: int = 9000):
        """Main loop for serving the site"""
        self.check_config()
        self.renderer_dict = {}
        observer = KartObserver(action=self.update_data)
        for miner in self.miners:
            miner.start_watching(self.config, observer)
        observer.start()
        for renderer in self.renderers:
            self.renderer_dict[renderer.name] = renderer
            renderer.start_serving(self.config)
        handler_class = KartRequestHandler
        handler_class.action = self.serve_page
        httpd = HTTPServer(("", port), handler_class)
        self.update_data()
        shutil.rmtree(self.build_location, ignore_errors=True)

        while True:
            try:
                httpd.handle_request()
            except KeyboardInterrupt:
                break
            except Exception:
                print(traceback.format_exc())

        print("\rexiting")
        for miner in self.miners:
            miner.stop_watching(self.config)
        for renderer in self.renderers:
            renderer.stop_serving(self.config)
        observer.stop()
        observer.join()

    def run(self):
        """Starts the kart execution. See --help for more information"""
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "command", help="command to execute", choices={"build", "serve"}
        )
        parser.add_argument(
            "-p", "--port", help="port to bind to", default=9000, type=int
        )
        parser.add_argument(
            "--dev-url",
            help="serve your site on a different url",
            type=str,
        )
        args = parser.parse_args()
        if args.command == "build":
            self.config["serving"] = False
            self.build()
        if args.command == "serve":
            self.config["serving"] = True
            if args.dev_url:
                self.config["site_url"] = args.dev_url
            else:
                self.config["site_url"] = f"http://localhost:{args.port}"
            self.serve(args.port)



from abc import ABC, abstractmethod
from pathlib import Path

from watchdog.events import RegexMatchingEventHandler

from kart.utils import id_from_path

try:
    from yaml import CSafeLoader as YamlLoader
except ImportError:
    from yaml import SafeLoader as YamlLoader
from typing import Dict

from kart.utils import KartDict, KartObserver
import mimetypes


class Miner(ABC):
    """Base miner class"""

    @abstractmethod
    def read_data(self, config: dict):
        """Reads data from its source"""

    @abstractmethod
    def collect(self, config: dict) -> Dict:
        """Collects all data"""

    @abstractmethod
    def start_watching(self, config: dict, observer: KartObserver):
        """Start watching for data changes"""

    @abstractmethod
    def stop_watching(self, config: dict):
        """Stop watching for data changes"""


class DefaultMiner(Miner):
    """Base miner class for reading from filesystem"""

    @abstractmethod
    def __init__(self):
        """Initializes miner. Must set the ``name`` and ``dir`` variables"""

    @abstractmethod
    def collect_single_file(self, file: Path, config: dict):
        """Reads data from a single file"""

    def read_data(self, config: dict):
        """Implements Miner.read_data().

        It iterates over a directory and calls collect_single_file() for each file
        """
        self.data = KartDict()
        for file in filter(Path.is_file, self.dir.iterdir()):
            if file.suffix not in self.extension:
                continue
            object = self.collect_single_file(file, config)
            if object:
                self.data.update(object)

    def collect(self, config: dict):
        return {self.name: self.data}

    def start_watching(self, config: dict, observer: KartObserver):
        """Registers a watchdog handler that calls collect_single_file() when a file has changed"""

        class Handler(RegexMatchingEventHandler):
            def on_moved(_, event):
                self.data.pop(id_from_path(self.dir, Path(event.src_path)))
                self.data.update(self.collect_single_file(Path(event.src_path), config))

            def on_modified(_, event):
                self.data.update(self.collect_single_file(Path(event.src_path), config))

            def on_deleted(_, event):
                self.data.pop(id_from_path(self.dir, Path(event.src_path)))

        self.read_data(config)
        observer.schedule(Handler(ignore_directories=True), self.dir, recursive=False)

    def stop_watching(self, config: dict):
        """Implements Miner.stop_watching(). It does nothing as no cleanup is needed"""


class DefaultMarkupMiner(DefaultMiner):
    """Base miner that implements collect_single_file() for markup files"""

    extension = (".md",)

    def collect_single_file(self, file: Path, config: dict) -> str:
        """
        Stores the data included in the frontmatter in a dictionary, adds
        the content in the ``content`` field and then return the dictionary
        """
        mimetypes.init()
        with file.open("r") as f:
            data = f.read().split("---")
            metadata = YamlLoader(data[1]).get_data()
            if "draft" in metadata and metadata["draft"] and not config["serving"]:
                return
            content = "---".join(data[2:])
            object = metadata
            slug = id_from_path(self.dir, file)
            object["slug"] = slug
            object["content"] = content
            object["content_type"] = mimetypes.guess_type(file)
            return {slug: object}


class DefaultCollectionMiner(DefaultMarkupMiner):
    """Miner that looks for data in the ``collections`` folder"""

    def __init__(self, collection: str, directory: str = "collections"):
        "Initializes miner. Sets the ``name``, ``dir`` and ``collection`` variables"
        self.collection = collection
        self.dir = Path() / directory / collection
        self.name = self.collection


class DefaultTaxonomyMiner(DefaultMarkupMiner):
    """Miner that looks for data in the ``taxonomy`` folder"""

    def __init__(self, taxonomy: str, directory: str = "taxonomies"):
        "Initializes miner. Sets the ``name``, ``dir`` and ``taxonomy`` variables"
        self.taxonomy = taxonomy
        self.dir = Path() / directory / taxonomy
        self.name = self.taxonomy


class DefaultPageMiner(DefaultMarkupMiner):
    """Miner that get data from the ``pages`` folder"""

    def __init__(self, directory: str = "pages"):
        "Initializes miner. Sets the ``name`` and ``dir`` variables"
        self.dir = Path(directory)
        self.name = "pages"


class DefaultDataMiner(DefaultMiner):
    """Miner that get yaml data from the ``data`` folder"""

    extension = (".yml", ".yaml")

    def __init__(self, directory: str = "data"):
        "Initializes miner. Sets the ``name`` and ``dir`` variables"
        self.dir = Path(directory)
        self.name = "data"

    def collect_single_file(self, file: Path, config: dict) -> str:
        """Users YamlLoader to get data from a yaml file"""
        with file.open("r") as f:
            slug = id_from_path(self.dir, file)
            return {slug: YamlLoader(f.read()).get_data()}



from pathlib import Path

from slugify import slugify
from watchdog.events import RegexMatchingEventHandler

from kart.miners import DefaultMarkupMiner
from kart.utils import KartDict

try:
    from yaml import CSafeLoader as YamlLoader
except ImportError:
    from yaml import SafeLoader as YamlLoader

import importlib
import inspect

import mistune
from jinja2 import contextfilter
from jinja2.runtime import Context
from mistune.directives import Directive

from kart.mappers import Mapper
from kart.ext.markdown import KartMistuneRenderer, TocRenderer
from kart.utils import KartObserver


class DefaultDocumentationMiner(DefaultMarkupMiner):
    """Miner that recursively looks for data in the ``docs`` folder"""

    def __init__(self, directory: str = "docs"):
        "Initializes miner. Sets the ``dir`` variable"
        self.dir = Path(directory)

    def __recursive_read_data(self, config: dict, dir: Path, level: int = 0):
        """Helper function"""
        if dir.joinpath("navigation.yml").exists():
            nav_file = dir.joinpath("navigation.yml").open()
            nav_data = YamlLoader(nav_file.read()).get_data()
            paths = []
            for x in nav_data:
                if "page" in x.keys():
                    paths.append(dir.joinpath(x["page"]))
                elif "section" in x.keys():
                    paths.append(dir.joinpath(x["section"]))
        else:
            paths = dir.iterdir()
            nav_data = []

        for i, item in enumerate(paths):
            if item.is_file():
                object = self.collect_single_file(item, config)
                slug, page = list(object.items())[0]
                toc_entry = {"title": page["title"], "slug": slug, "level": level}
                self.docs_global_toc.append(toc_entry)
                if object:
                    self.markdown_data.update(object)
            elif item.is_dir():
                toc_entry = {"title": nav_data[i]["name"], "slug": None, "level": level}
                self.docs_global_toc.append(toc_entry)
                self.__recursive_read_data(config, item, level + 1)

    def read_data(self, config: dict):
        self.markdown_data = KartDict()
        self.docs_global_toc = []
        self.__recursive_read_data(config, self.dir)

    def collect(self, config: dict):
        return {"docs": self.markdown_data, "docs_global_toc": self.docs_global_toc}

    def start_watching(self, config: dict, observer: KartObserver):
        """Registers a watchdog handler that calls read_data() when a file has changed"""

        class Handler(RegexMatchingEventHandler):
            def on_any_event(_, event):
                self.read_data()

        self.read_data(config)
        observer.schedule(Handler(), self.dir, recursive=True)


class DefaultDocumentationMapper(Mapper):
    """Mapper intended to be used with DefaultDocumentationMapper"""

    def __init__(self, template: str = "page.html", base_url: str = ""):
        self.template = template
        self.base_url = base_url

    def map(self, config: dict, site: KartDict) -> dict:
        urls = {}
        previous_slug = None
        for slug, page in site["docs"].items():
            if "url" in page:
                url = page["url"]
            elif slug == "index":
                url = "/"
            else:
                url = "/" + "/".join(slugify(part) for part in slug.split(".")) + "/"
            if "template" in page:
                template = page["template"]
            else:
                template = self.template
            if len(urls):
                page["previous_page"] = previous_slug
                urls[previous_slug]["data"]["next_page"] = slug
            previous_slug = slug
            map_page = {
                "url": self.base_url + url,
                "data": {**page},
                "template": template,
                "renderer": "default_site_renderer",
            }
            urls[slug] = map_page
        return urls


class DocumentationDirective(Directive):
    """Mistune class that add the ``function`` and ``class`` directive to build a techical documentation"""

    def parse(self, block, m, state):
        name = m.group("name")
        title = m.group("value")
        text = self.parse_text(m)
        children = block.parse(text, state, block.rules)
        return {"type": name, "children": children, "params": (name, title)}

    def render_html_function(self, text, name, loc):
        """Renders the ``function`` directive"""
        module_name = ".".join(loc.split(".")[:-1])
        func_name = loc.split(".")[-1]
        module = importlib.import_module(module_name)
        module = importlib.reload(module)
        func = module.__dict__[func_name]
        sig = inspect.signature(func)
        html = "<dl>"
        html += f'<dt id="{loc}">function {loc}{sig}</dt>'
        html += f"<dd><p>{func.__doc__}</p></dd>"
        html += "</dl>"
        return html

    def render_html_class(self, text, name, loc):
        """Renders the ``class`` directive"""
        module_name = ".".join(loc.split(".")[:-1])
        func_name = loc.split(".")[-1]
        module = importlib.import_module(module_name)
        module = importlib.reload(module)
        cls = module.__dict__[func_name]
        parents = []
        for p in cls.__bases__:
            parents.append(p.__module__ + "." + p.__name__)
        html = '<dl classs="class">'
        html += f'<dt id="{loc}">class {loc}({", ".join(parents)})</dt>'
        html += f"<dd><p>{cls.__doc__}</p></dd>"
        functions = []
        for x in inspect.getmembers(cls):
            try:
                if x[1].__module__ != module_name:
                    continue
                if x[1].__qualname__.split(".")[0] != cls.__name__:
                    continue
            except Exception:
                continue
            functions.append(cls.__dict__[x[0]])
        if functions:
            html += "<dl>"
        for func in functions:
            sig = inspect.signature(func)
            if inspect.isabstract(cls) and func.__name__ in cls.__abstractmethods__:
                html += "<dt>@abstractmethod</dt>"
            html += f"<dt>{func.__name__}{sig}</dt>"
            if inspect.getdoc(func):
                html += f"<dd><p>{inspect.getdoc(func)}</p></dd>"
        if functions:
            html += "</dl>"

        html += "</dl>"
        return html

    def render_ast(self, children, name, title=None):
        return {"type": name, "children": children, "name": name, "title": title}

    def __call__(self, md):
        for name in {"function", "class"}:
            self.register_directive(md, name)
            if md.renderer.NAME == "ast":
                md.renderer.register(name, self.render_ast)

        if md.renderer.NAME == "html":
            md.renderer.register("function", self.render_html_function)
            md.renderer.register("class", self.render_html_class)


@contextfilter
def markdown_to_html(context: Context, markdown: str) -> str:
    """Converts markdown data to html.
    It supports markdown directives to extract the documentation out of python
    docstrings.
    """
    return mistune.Markdown(
        renderer=KartMistuneRenderer(context=context, escape=False),
        plugins=[DocumentationDirective()],
    )(markdown)


class DocumentationTocRenderer(TocRenderer):
    """Mistune renderer used by markdown_to_toc()"""

    def __init__(self):
        self._methods = {"class": self._class}
        self.SUPPORTED_ELEMENTS = {"heading", "text", "class", "function"}

    def function(self, _, type, name, **kwargs):
        """Renders the ``function`` directive"""
        return {"title": name.split(".")[-1], "id": name, "level": 2}

    def _class(self, _, type, name, **kwargs):
        """Renders the ``class`` directive"""
        return {"title": name.split(".")[-1], "id": name, "level": 2}


def markdown_to_toc(markdown: str) -> str:
    """Extracts a list of header from markdown data"""
    return mistune.Markdown(
        renderer=DocumentationTocRenderer(),
        plugins=[DocumentationDirective()],
    )(markdown)



from abc import ABC, abstractmethod
from datetime import datetime, time
from distutils.dir_util import copy_tree
from http.server import SimpleHTTPRequestHandler
from multiprocessing import Pool
from pathlib import Path

from dateutil import tz
from jinja2 import Environment, FileSystemLoader

from kart.ext.markdown import markdown_to_html, markdown_to_toc
from kart.utils import KartDict, KartMap, date_to_string


class Renderer(ABC):
    """Base Renderer class"""

    @abstractmethod
    def render(self, config: dict, site: KartDict, map: KartMap, build_location: str):
        """Renders each page indented for this renderer"""

    def start_serving(self, config: dict):
        """Start the dev server, if necessary"""

    @abstractmethod
    def serve(
        self, http_handler, page: dict, config: dict, site: KartDict, map: KartMap
    ):
        """Serves the the requested page"""

    def stop_serving(self, config: dict):
        """Stops the dev server, if necessary"""


class DefaultFileRenderer(Renderer):
    """Base class for renderers that render file individually"""

    @abstractmethod
    def __init__(self, name: str):
        """Initializes renderer. Must set the ``name`` and ``content_type`` variables"""

    @abstractmethod
    def render_single(
        self, page: dict, config: dict, site: KartDict, map: KartMap
    ) -> str:
        """Renders a single file"""

    def render(self, config, site, map, build_location):
        for page in map:
            if page["renderer"] != self.name:
                continue
            rendered_file = self.render_single(page, config, site, map)
            path = Path(build_location) / Path(*Path(page["url"]).parts[1:])
            path.parent.mkdir(parents=True, exist_ok=True)
            with path.open("w") as f:
                f.write(rendered_file)

    def serve(
        self, http_handler, page: dict, config: dict, site: KartDict, map: KartMap
    ):
        http_handler.send_response(200)
        http_handler.send_header("Content-type", self.content_type)
        http_handler.end_headers()
        content = self.render_single(page, config, site, map)
        http_handler.wfile.write(bytes(content, "utf-8"))


class DefaultDirectoryRenderer(Renderer):
    """Base class for renderers that render directories"""

    @abstractmethod
    def __init__(self, name: str, directory: str):
        """Initializes renderer. Must set the ``name``, ``dir`` and ``base_url`` variables"""

    def render(self, config: dict, site: KartDict, map: KartMap, build_location: str):
        """Copies the entire directory to the target destination"""
        for page in map:
            if page["renderer"] != self.name:
                continue
            destination = Path(build_location) / Path(*Path(page["url"][:-1]).parts[1:])
            copy_tree(self.dir, str(destination))

    def serve(
        self, http_handler, page: dict, config: dict, site: KartDict, map: KartMap
    ):
        http_handler.path = self.base_url + http_handler.path
        return SimpleHTTPRequestHandler.do_GET(http_handler)


class DefaultSiteRenderer(DefaultFileRenderer):
    """Default renderer for rendering html files"""

    def __init__(
        self,
        name: str = "default_site_renderer",
        template_folder: str = "templates",
        filters: dict = {
            "html": markdown_to_html,
            "toc": markdown_to_toc,
            "date_to_string": date_to_string,
        },
        process_count: int = 1,
    ):
        self.name = name
        self.content_type = "text/html"
        self.template_folder = template_folder
        self.process_count = process_count
        self.env = Environment(loader=FileSystemLoader(self.template_folder))
        self.env.filters.update(filters)

    def render_single(
        self, page: dict, config: dict, site: KartDict, map: KartMap
    ) -> str:
        template = self.env.get_template(page["template"])
        page = {**page["data"], "url": page["url"]}
        return template.render(page=page, config=config, site=site, url=map.url)

    def _render_single(
        self, key: str, config: dict, site: KartDict, map: KartMap, build_location: str
    ):
        """Renderers a single file and saves it to the disk"""
        page = map[key]
        if page["renderer"] != self.name:
            return
        rendered_file = self.render_single(page, config, site, map)
        path = Path(build_location) / Path(*Path(page["url"]).parts[1:])
        path = path / "index.html"
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("w") as f:
            f.write(rendered_file)

    def render(
        self, config: dict, site: KartDict, map: KartMap, build_location: str = "_site"
    ):
        """Renderers all the files with a multiprocessing Pool for faster build times"""
        if self.process_count == 1:
            for key in map.keys():
                self._render_single(key, config, site, map, build_location)
        else:
            with Pool(self.process_count) as p:
                data = ((key, config, site, map, build_location) for key in map.keys())
                p.starmap(self._render_single, data)


class DefaultFeedRenderer(DefaultFileRenderer):
    """Renders an atom feed file"""

    def __init__(self, name: str = "default_feed_renderer"):
        """Initializes renderer. Sets the ``name`` and ``content_type`` variables"""
        self.name = name
        self.content_type = "application/xml"

    def render_single(
        self, page: dict, config: dict, site: KartDict, map: KartMap
    ) -> str:
        """Creates the atom feeds"""
        feed_entries = []
        for collection in page["data"]["collections"]:
            for object in site[collection]:
                feed_entries.append([map.url(collection, object["slug"]), object])
        feed_entries.sort(key=lambda x: x[1]["date"], reverse=True)
        atom = '<feed xmlns="http://www.w3.org/2005/Atom">'
        atom += f'<id>"{map.url("/")}"</id>'
        atom += f"<title>{config['name']}</title>"
        timezone = config["timezone"]
        updated_time = datetime.now().replace(tzinfo=tz.gettz(timezone))
        atom += f"<updated>{updated_time.isoformat()}</updated>"
        atom += f'<link href="{map.url("/")}"/>'
        atom += f'<link href="{map.url(page["url"])}" rel="self"/>'
        for url, entry in feed_entries:
            atom += "<entry>"
            atom += f"<id>{url}</id>"
            title = entry["title"] if "title" in entry.keys() else entry["name"]
            atom += f"<title>{title}</title>"
            entry_time = datetime.combine(entry["date"], time(12))
            entry_time.replace(tzinfo=tz.gettz(timezone))
            atom += f"<updated>{entry_time.isoformat()}</updated>"
            if "description" in entry.keys():
                atom += f'<summary>{entry["description"]}</summary>'
            atom += f'<link href="{url}" link="alternate"/>'
            atom += "</entry>"
        atom += "</feed>"
        return atom


class DefaultSitemapRenderer(DefaultFileRenderer):
    """Renders the sitemap of the site"""

    def __init__(self, name: str = "default_sitemap_renderer"):
        """Initializes renderer. Sets the ``name`` and ``content_type`` variables"""
        self.name = name
        self.content_type = "application/xml"

    def render_single(self, page: dict, config: dict, site: KartDict, map: KartMap):
        """Uses the site ``map`` variable to create the sitemap"""
        sitemap = '<?xml version="1.0" encoding="UTF-8"?>'
        sitemap += '<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">'
        for x in map:
            if x["renderer"] != "default_site_renderer":
                continue
            sitemap += f"<url><loc>{map.url(x['url'])}</loc></url>"
        sitemap += "</urlset>"
        return sitemap


class DefaultStaticFilesRenderer(DefaultDirectoryRenderer):
    """Renders all files in the ``static`` directory"""

    def __init__(
        self, name: str = "default_static_files_renderer", directory: str = "static"
    ):
        """Initializes renderer. Sets the ``name``, ``dir`` and ``content_type`` variables"""
        self.name = name
        self.dir = directory
        self.base_url = ""


class DefaultRootDirRenderer(DefaultDirectoryRenderer):
    """Renders all files in the ``root`` directory"""

    def __init__(
        self, name: str = "default_root_dir_renderer", directory: str = "root"
    ):
        """Initializes renderer. Sets the ``name``, ``dir`` and ``content_type`` variables"""
        self.name = name
        self.dir = directory
        self.base_url = "/root"



from jinja2 import contextfilter
from jinja2.runtime import Context
from subprocess import Popen, PIPE
import xml.etree.ElementTree as xml


@contextfilter
def asciidoc_to_html(context: Context, asciidoc: str) -> str:
    """Converts asciidoc data to html"""
    parsed_asciidoc = context.environment.from_string(asciidoc).render(context)
    config = context["config"]["code_highlighting"]
    pygments_header = ":source-highlighter: pygments\n"
    if config["noclasses"]:
        pygments_header += ":pygments-css: style\n"
        pygments_header += f':pygments-style: {config["style"]}\n'
    parsed_asciidoc = pygments_header + parsed_asciidoc
    p = Popen(["asciidoctor", "-e", "-"], stdin=PIPE, stdout=PIPE)
    return p.communicate(input=parsed_asciidoc.encode("utf-8"))[0].decode("utf-8")


def asciidoc_to_toc(asciidoc: str) -> str:
    """Extracts a list of header from asciidoc data"""
    asciidoc = ":toc:\n:toclevels: 4\n" + asciidoc
    p = Popen(["asciidoctor", "-e", "-a", "toc", "-"], stdin=PIPE, stdout=PIPE)
    html = p.communicate(input=asciidoc.encode("utf-8"))[0].decode("utf-8")
    toc_xml = xml.fromstring("</div>".join(html.split("</div>")[:2]) + "</div>")[1]

    def parse(ul):
        toc = []
        for li in ul:
            for x in li:
                if x.tag == "a":
                    toc.append(
                        {
                            "title": x.text,
                            "id": x.attrib["href"][1:],
                            "level": int(ul.attrib["class"][-1:]) + 1,
                        }
                    )
                elif x.tag == "ul":
                    toc.extend(parse(x))
        return toc

    return parse(toc_xml)



import math
import queue
import traceback
from collections import OrderedDict
from datetime import datetime
from http.server import SimpleHTTPRequestHandler
from pathlib import Path
from typing import Callable, List

from watchdog.observers import Observer


class KartObserver(Observer):
    """Extends whatchdog observer to execute a function after each event"""

    def __init__(self, action: Callable):
        """Initializes the observer with the ``action`` variable"""
        super().__init__()
        self.action = action

    def run(self):
        """Ovverrides Observer.run() to execute an function each time an even is handled"""
        while self.should_keep_running():
            try:
                self.dispatch_events(self.event_queue, self.timeout)
                if self.event_queue.empty():
                    self.action()
            except queue.Empty:
                pass
            except Exception:
                print(traceback.format_exc())


class KartRequestHandler(SimpleHTTPRequestHandler):
    """Extends SimpleHTTPRequestHandler execute a function for each request"""

    def do_GET(self):
        """Executes self.action() for every get request"""
        self.action(self, self.path)


class KartDict(OrderedDict):
    """Subclass of OrderedDict with a custom iterator"""

    def __iter__(self):
        """Iterate over the dict values instead of the keys"""
        return iter(self.values())


class KartMap(KartDict):
    """Custom dictionary that holds the site map"""

    def __init__(self, initial_data: dict = {}, site_url: str = "", *args, **kwargs):
        super().__init__(initial_data, *args, **kwargs)
        self.site_url = site_url

    def url(self, *name: List[str]) -> str:
        """Takes as input the slug of a page and returns its absolute url"""
        name = ".".join(name)
        if not name:
            return ""
        if name in self.keys():
            return self.site_url + self[name]["url"]
        elif name + ".1" in self.keys():
            return self.site_url + self[name + ".1"]["url"]
        elif "http" in name:
            return name
        elif "/" in name:
            return self.site_url + name
        print(f'Invalid url: "{name}"')
        return ""


def paginate(
    objects: list,
    per_page: int,
    template: str,
    base_url: str,
    slug: str,
    renderer: str,
    additional_data: dict = {},
) -> dict:
    """Splits objects and for each group create an entry in the site map"""
    urls = {}
    paginated_objects = [
        objects[x * per_page : (x + 1) * per_page]
        for x in range(max(math.ceil(len(objects) / per_page), 1))
    ]
    for i, objects in enumerate(paginated_objects, 1):

        if i > 1:
            previous_page = f"{slug}.{i-1}"
        else:
            previous_page = ""

        if i < len(paginated_objects):
            next_page = f"{slug}.{i+1}"
        else:
            next_page = ""

        paginator = {
            "objects": objects,
            "index": i,
            "next_page": next_page,
            "previous_page": previous_page,
        }
        data = {"paginator": paginator}
        data.update(additional_data)
        urls[f"{slug}.{i}"] = {
            "url": base_url + f"{i}/" if i > 1 else base_url,
            "data": data,
            "template": template,
            "renderer": renderer,
        }
    return urls


def date_to_string(date: datetime) -> str:
    "Formats a date to be displayed"
    return date.strftime("%b %d, %Y")


def id_from_path(base_dir: Path, path: Path) -> str:
    "Return the identifier of a page from its relative filepath"
    path = path.relative_to(base_dir)
    id = ".".join(path.parts)[: -len(path.suffix)]
    return id


def merge_dicts(a: dict, b: dict) -> dict:
    """Merge two dicts"""
    for key in b:
        if key not in a:
            a[key] = b[key]
        else:
            if isinstance(b[key], dict):
                if isinstance(a[key], dict):
                    merge_dicts(a[key], b[key])
                else:
                    raise Exception
